#!/usr/bin/env python3
"""
Module SOCKETWRAPPER -- Socket Pair Wrapper Function
Sub-Package COMM of Package PLIB3 -- General Python Utilities
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the ``socketpair_wrapper`` function, which
provides a portable interface for creating socket pairs to
communicate with child processes and threads.

Note that on Windows the ``multiprocessing`` module is used
for child process support, which is only available in Python
2.6 and later.
"""

import socket


if hasattr(socket, 'socketpair'):
    # We're on a Unix-type system, so no special gymnastics needed
    
    
    def _child(client_sock, daemon_sock, fn, is_process):
        # Child process/thread
        
        # Hack so we can use the same function for both
        # processes and threads (in a child thread we don't
        # want to close the client socket since that will
        # kill it for the parent thread too)
        if is_process:
            try:
                client_sock.close()
            except:
                return 248
        try:
            # The user function is responsible for closing the socket
            # (this includes if an exception happens--the except
            # clause below is only meant to catch exceptions that
            # the user function doesn't want to define a different
            # return code for)
            retcode = fn(daemon_sock)
            if retcode is None:
                retcode = 0
            return retcode
        except:
            return 249
    
    
    def socketpair_wrapper(fn, Wrapper):
        is_process = 'Process' in Wrapper.__name__
        client_sock, daemon_sock = socket.socketpair()
        child = Wrapper(_child, client_sock, daemon_sock, fn,
                        is_process)
        child.start()
        # Same hack as above, don't want to close the daemon socket
        # in a thread since that kills it for the child too
        if is_process:
            daemon_sock.close()
        # The caller is responsible for closing the client socket
        return (client_sock, child)


else:
    # We're on Windows, so we have to emulate socketpair
    
    
    def _child(lsock, port, fn, is_process):
        # Child process/thread, must connect with parent to establish
        # the socket pair
        if is_process:
            try:
                lsock.close()
            except:
                return 245
        try:
            daemon_sock = socket.socket()
        except:
            return 246
        try:
            daemon_sock.connect(('localhost', port))
        except:
            return 247
        try:
            # The user function is responsible for closing the socket
            # (this includes if an exception happens--the except
            # clause below is only meant to catch exceptions that
            # the user function doesn't want to define a different
            # return code for)
            retcode = fn(daemon_sock)
            if retcode is None:
                retcode = 0
            return retcode
        except:
            return 249
    
    
    def socketpair_wrapper(fn, Wrapper):
        is_process = 'Process' in Wrapper.__name__
        listensock = socket.socket()
        listensock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        listensock.bind(('localhost', 0))
        iface, ephport = listensock.getsockname()
        listensock.listen(1)
        child = Wrapper(_child, listensock, ephport, fn, is_process)
        child.start()
        client_sock, addr = listensock.accept()
        listensock.close()
        # The caller is responsible for closing the client socket
        return (client_sock, child)



#!/usr/bin/env python3
"""
Module SelfPipe
Sub-Package COMM.CLASSES of Package PLIB3
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the ``SelfPipe`` class, which implements
the self-pipe trick in a general way that can be used by any
application wanting to multiplex socket I/O with signals.
"""


from plib.comm import sigtools


class SelfPipe(object):
    """Self-pipe manager.
    
    Wraps the low-level API provided by the ``sigtools``
    module, which uses the ``set_wakeup_fd`` mechanism in
    the Python ``signal`` module.
    """
    
    def __init__(self, callback):
        self.pipe_fd = sigtools.selfpipe_fd()
        self.callback = callback
        self.old_handlers = {}
    
    def fileno(self):
        return self.pipe_fd
    
    def track_signal(self, sig, reset=False):
        self.old_handlers[sig] = sigtools.track_signal(sig, reset)
    
    def track_signals(self, *sigs):
        for sig, handler in zip(sigs, sigtools.track_signals(*sigs)):
            self.old_handlers[sig] = handler
    
    def send_signal(self, sig):
        sigtools.send_signal(sig)
    
    def receive_signals(self):
        for sig in sigtools.signals_received():
            self.callback(sig)



#!/usr/bin/env python3
"""
Module SigChldMixin
Sub-Package COMM.CLASSES of Package PLIB3
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the ``SigChldMixin class``, which allows
parent processes to respond to SIGCHLD by reaping any
child processes that have completed. The code is
general enough that it should be useful for any process
that wants to fork children and reap them properly. See
the method docstrings for notes on the particular strategy
chosen in this class for handling SIGCHLD. Note that this
functionality is only useful on platforms (e.g., *not*
Windows) that have SIGCHLD; however, the other methods of
the class, which actually do the tracking and reaping of
children, can be used on all platforms provided that some
other mechanism is used to trigger them.
"""

import signal


class SigChldMixin(object):
    """Mixin class for parent processes to properly reap dead children.
    
    This class implements logic to reap zombie children when they
    exit. The logic here is general; for a mixin that is specifically
    designed for use with PLIB3 servers, see ``SigChldServerMixin``.
    The key requirement is that the ``track_child`` method must be
    called on each child instance when it is constructed (the
    ``SigChldServerMixin`` class does this in the ``_new_child``
    method, right before the new child object is returned). The
    only assumption made is that each ``child`` is an instance of
    the ``ProcessWrapper`` class in ``plib.comm._processwrapper``,
    since that API is used to interact with children.
    """
    
    active_children = None
    reaping = False
    
    def setup_child_sig_handler(self):
        """Set up the SIGCHLD handler.
        """
        if hasattr(signal, 'SIGCHLD'):
            signal.signal(signal.SIGCHLD, self.child_sig_handler)
    
    def child_sig_handler(self, sig, frame=None):
        # Reap children
        self.reap_children()
        # Ensure BSD semantics
        signal.signal(signal.SIGCHLD, self.child_sig_handler)
    
    def track_child(self, child):
        """Start tracking a new child process.
        """
        
        if self.active_children is None:
            self.active_children = []
        self.active_children.append(child)
    
    def check_child(self, child):
        # Must override to return true if child has exited
        raise NotImplementedError
    
    def _wait_children(self):
        # Internal routine to wait for children that have exited.
        if self.active_children is None:
            return
        
        # XXX: This loop runs more system calls than it ought
        # to. There should be a way to put the active_children into a
        # process group and then use os.waitpid(-pgid) to wait for any
        # of that set, but I couldn't find a way to allocate pgids
        # that couldn't collide.
        removes = []
        for child in self.active_children:
            if self.check_child(child):
                removes.append(child)
        for child in removes:
            self.active_children.remove(child)
    
    def reap_children(self):
        """Reap children if any are waiting.
        
        If we received SIGCHLD, it will break us out of any pending system
        call; this method is intended to be called before re-starting the
        system call to ensure that there are no zombies remaining.
        
        You may ask: why not just set the SA_RESTART flag when we call
        sigaction to set up our SIGCHLD handler? Then we could just reap
        children inside the handler without affecting the logic of our
        main program loop. There are several reasons why we don't do that
        in this library. First, the select system call is not supposed
        to be restartable in this way, and many of the use cases for
        this class will likely involve select (this includes adding
        SIGCHLD handling to the forking socket server in the PLIB3 I/O
        sub-package, since that server's main loop uses select in order
        to enable the self-pipe trick for termination signals). Second,
        even for system calls which are restartable (e.g., the accept
        call, which a forking server that didn't use the self-pipe trick
        could use as its main "waiting" call), the SA_RESTART flag is
        not guaranteed to produce the desired semantics on all platforms.
        Third, the timing of Python signal handlers when a signal arrives
        during a system call is not the same as the timing of a signal
        handler in a C program would be. A C signal handler runs at
        interrupt time--i.e., as soon as the signal is received. But a
        Python signal handler does not; at the C level, the Python
        interpreter just sets a flag inside its signal handler and
        returns. The Python handler, if one is installed, does not run
        until the interpreter can pause between bytecodes to switch
        context and run it--which it can't do while a system call is in
        progress, because the interpreter doesn't have control, the
        kernel does. This means that, as far as the kernel-level system
        call is concerned, the "signal handler" is actually finished when
        the C code inside the interpreter sets the flag--which means that
        even if the SA_RESTART flag were set, it would not cause the
        *Python* signal handler to run when we want it to run, which is
        soon after the signal arrives. Instead, the Python handler would
        not run until some *other* event caused the pending system call,
        which was restarted by the kernel because SA_RESTART was set, to
        return! So for Python signal handlers, our only choice is to let
        the signal interrupt the system call, which will then return, at
        the C level, with an EINTR error; the interpreter then checks
        the signal flag, sees that a signal was received, and runs our
        Python signal handler when we want it run, right after the signal
        arrived. It's true that whatever Python code originally made the
        system call now has to know to restart the call if an EINTR error
        occurs, but that code has to know that anyway, because it has to
        know to check the flag that our Python signal handler set, so it
        can reap children.
        """
        
        if not self.reaping:
            self.reaping = True
            self._wait_children()
            self.reaping = False
    
    def end_child(self, child):
        # Must override to shut down child
        raise NotImplementedError
    
    def close_children(self):
        """Close all active children unconditionally.
        
        Intended to be called when the parent is shutting down.
        """
        
        # This keeps us from being interrupted by SIGCHLD while we're
        # closing down the children by hand (on some platforms, e.g.
        # OS X, the parent can be interrupted by SIGCHLD when a child
        # exits even if it's already inside a waitpid system call on
        # the *same* child)
        if hasattr(signal, 'SIGCHLD'):
            signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        if self.active_children is not None:
            for child in self.active_children:
                self.end_child(child)
            del self.active_children



#!/usr/bin/env python3
"""
Module WAITWRAPPER -- Child Waiting Wrapper Function
Sub-Package COMM of Package PLIB3 -- General Python Utilities
Copyright (C) 2008-2015 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the ``wait_wrapper`` function, which
starts a child process or thread and then waits until the child
has started before continuing.

Note that on Windows the ``multiprocessing`` module is used
for child process support, which is only available in Python
2.6 and later.
"""

from os import name as os_name

server_ok = b"1"

if os_name == "posix":
    # We're on a Unix-type system and can use os.pipe safely
    
    import os
    
    
    def _child(rfd, wfd, start_fn, run_fn, is_process):
        # Child process/thread; note that the exit codes we return
        # are out of the range of possible errno values, to ensure
        # that we can tell what error happened
        try:
            try:
                # Hack so we can use the same function for both
                # processes and threads (in a child thread we don't
                # want to close this file descriptor since that will
                # kill it for the parent thread too)
                if is_process:
                    os.close(rfd)
                start_fn()
                os.write(wfd, server_ok)
            except:
                return 253
        finally:
            try:
                # We can close this file descriptor since we're done
                # with it
                os.close(wfd)
            except:
                pass
        try:
            retcode = run_fn()
            if retcode is None:
                retcode = 0
            return retcode
        except:
            return 254
    
    
    def wait_wrapper(start_fn, run_fn, Wrapper):
        is_process = 'Process' in Wrapper.__name__
        rfd, wfd = os.pipe()
        child = Wrapper(_child, rfd, wfd, start_fn, run_fn,
                        is_process)
        child.start()
        # Same hack as above, don't want to close this file descriptor
        # in a thread since that kills it for the child too
        if is_process:
            os.close(wfd)
        try:
            ok = os.read(rfd, len(server_ok))
            if ok != server_ok:
                raise IOError("fork_wait aborted!")
        finally:
            # We can close this file descriptor since we're done
            # with it
            os.close(rfd)
        return child


else:
    # We're on Windows so os.pipe won't work as is; we could use the
    # multiprocessing module's Pipe object (which allows pipes to be
    # inherited by child processes), but it doesn't seem to work with
    # the fork_wait logic we're using (closing one end of the pipe
    # seems to guarantee EOFError on the other end even if data was
    # written before the close and therefore should be readable--???),
    # so we use socketpair_wrapper instead, which is known to work
    
    from functools import partial
    
    from ._socketwrapper import socketpair_wrapper
    
    
    def _child(start_fn, run_fn, sock):
        # We'll form a partial so the function sent to
        # socketpair_wrapper only takes the sock argument (this
        # works as long as we use functools.partial directly,
        # to ensure that it's picklable by multiprocessing)
        try:
            try:
                start_fn()
                sock.sendall(server_ok)
            except:
                return 253
        finally:
            try:
                # We can close this socket here since we're done with it
                sock.close()
            except:
                pass
        try:
            retcode = run_fn()
            if retcode is None:
                retcode = 0
            return retcode
        except:
            return 254
    
    
    def wait_wrapper(start_fn, run_fn, Wrapper):
        sock, child = socketpair_wrapper(
            partial(_child, start_fn, run_fn),
            Wrapper)
        try:
            ok = sock.recv(len(server_ok))
            if ok != server_ok:
                raise IOError("fork_wait aborted!")
        finally:
            # We can close this socket here since we're done with it
            sock.close()
        return child



#!/usr/bin/env python3
"""
Module CHILDWRAPPER -- Base Child Process/Thread Wrapper Object
Sub-Package COMM of Package PLIB3 -- General Python Utilities
Copyright (C) 2008-2015 by Peter A. Donis

This module contains the ``ChildWrapper`` class, which provides
a portable abstract interface to child processes and threads
and their exit status.
"""


class ChildWrapper(object):
    """Base abstract interface for handling child processes/threads.
    
    The ``shutdown_with_parent`` flag determines whether the
    child should be shut down when its parent finishes, or
    whether it should continue independently. The former is
    typical for, say, request handlers spawned by servers;
    the latter is more useful for, say, forking the server
    process itself. The default is ``False``.
    """
    
    shutdown_with_parent = False
    
    _exitcode = None
    
    def __init__(self, target, *args, **kwargs):
        self._target = target
        self._args = args
        self._kwargs = kwargs
    
    def start(self):
        """Start the child process/thread.
        
        This method must only start the child once; if it is
        called multiple times, the second and subsequent calls
        must do nothing to the process. It is undefined whether
        multiple calls to this method will raise an exception
        (the implementations below raise RuntimeError).
        """
        raise NotImplementedError
    
    def stop(self):
        """Stop the child process/thread.
        
        This method must not cause an error if called multiple times.
        Thus, it must be able to tell if the process has already been
        stopped, and not try to stop it again.
        """
        raise NotImplementedError
    
    def check(self):
        """Check whether the child process/thread has stopped.
        
        If this function returns true, ``exitcode`` can be called to
        get the child's exit code.
        """
        raise NotImplementedError
    
    def wait(self):
        """Wait for the child process/thread to exit and get its exit code.
        
        After this method is called, ``self._exitcode`` must contain
        the exit code of the process.
        
        This method must not cause an error if called multiple times.
        Thus, it must be able to tell if the process has already
        exited, and not try to wait for it again.
        """
        raise NotImplementedError
    
    def exitcode(self):
        """Return the child process/thread exit code.
        
        Note that this method does not force the child to stop, so
        it could hang indefinitely unless ``stop`` is called before
        this method (or unless you have some other way of ensuring
        that the child is stopped first).
        """
        self.wait()
        return self._exitcode
    
    def end(self):
        """Ensure that the child is done.
        
        The exit code can be retrieved by calling ``exitcode``
        after this method returns.
        """
        self.stop()
        self.wait()



#!/usr/bin/env python3
"""
Module SERVERPROXY -- Server Proxy Object
Sub-Package COMM of Package PLIB3 -- General Python Utilities
Copyright (C) 2008-2015 by Peter A. Donis

This module contains the ``ServerProxy`` class, which is used
by the ``fork_server`` function in this sub-package to create
child processes that run servers. This class is separated into
its own module so that the child process that uses it does not
need to import the ``forkserver`` module or its dependencies
(this really only makes a difference on Windows, where the
``multiprocessing`` module is used to run the child process).
"""


class ServerProxy(object):
    
    def __init__(self, server_class, server_addr=None, handler_class=None):
        self.server_class = server_class
        self.server_addr = server_addr
        self.handler_class = handler_class
    
    def start_server(self):
        # Allow a tuple of (<module_name>, <class_name>) for the server
        # and handler classes so they can be imported only in the child
        # process, to reduce the memory footprint prior to forking
        for attrname in ('server_class', 'handler_class'):
            obj = getattr(self, attrname)
            if isinstance(obj, tuple):
                from plib.stdlib.imp import import_from_module
                setattr(self, attrname, import_from_module(*obj))
        if self.handler_class is not None:
            self.server = self.server_class(
                self.server_addr, self.handler_class)
        elif self.server_addr is not None:
            self.server = self.server_class(
                self.server_addr)
        else:
            self.server = self.server_class()
    
    def run_server(self):
        self.server.serve_forever()
